@@ -85,7 +85,7 @@ export const createShortLink = memoizeOne(async (path: string): Promise<string>
  * @param path - The long path to share.
  * @returns A ClipboardItem for the shortened link.
  */
-const createShortLinkClipboardItem = (path: string) => {
+export const createShortLinkClipboardItem = (path: string) => {
   return new ClipboardItem({
     'text/plain': createShortLink(path),
   });
